import typing as t
from frozendict import frozendict


class i18nstr(str):

    __slots__ = ('domain', 'context', 'default', 'mapping')

    def __new__(self,
                msgid: t.Union[str, 'i18nstr'],
                domain: t.Optional[str] = None,
                default: t.Optional[str] = None,
                mapping: t.Optional[t.Mapping] = None,
                context: t.Optional[t.Any] = None):

        self = super().__new__(self, str(msgid))
        if isinstance(msgid, self.__class__):
            domain = domain or msgid.domain
            context = context or msgid.context
            default = default or msgid.default
            mapping = mapping or msgid.mapping
        self.domain = domain
        self.context = context
        self.default = default or str(msgid)
        self.mapping = mapping and frozendict(mapping) or None
        return self

    def replace(self, **kwargs) -> 'i18nstr':
        return self.__class__(self, **kwargs)



import polib
import re
import logging
import typing as t
from pathlib import Path
from langcodes import Language
from langcodes.tag_parser import LanguageTagError


Logger = logging.getLogger(__name__)


def compiled(pofile: Path, mofile: Path):
    """Creates or updates a mo file in the locales folder.
    """
    try:
        po_mtime = pofile.stat().st_mtime
    except (IOError, OSError):
        # please log.
        return

    if mofile.exists():
        # Update mo file?
        try:
            mo_mtime = mofile.stat().st_mtime
        except (IOError, OSError):
            # please log.
            return
    else:
        mo_mtime = 0

    if po_mtime > mo_mtime:
        try:
            po = polib.pofile(pofile)
            po.save_as_mofile(mofile)
        except (IOError, OSError) as e:
            logging.warn('Error while compiling %s (%s).' % (pofile, e))
            raise

    return mofile


def iter_translation_sources(root: Path):
    for langtag in root.iterdir():
        if langtag.is_dir():
            messages_dir = langtag / 'LC_MESSAGES'
            try:
                tag = Language.get(langtag.stem).to_tag()
            except LanguageTagError:
                Logger.warning(f'{langtag.stem} is not a valid language.')
            for filename in messages_dir.iterdir():
                if filename.suffix == '.po':
                    yield tag, filename


def iter_translation_files(root: Path, can_compile: bool = False):
    """Expects a classical gettext directory structure:
        {root}/{langtag}/LC_MESSAGES/{domain}.mo
    """
    for tag, pofile in iter_translation_sources(root):
        mofile = pofile.with_suffix('.mo')
        if can_compile:
            if compiled(pofile, mofile) == mofile:
                yield tag, mofile
        elif mofile.exists():
            yield tag, mofile
        else:
            Logger.warning(
                f'File {pofile} does not have a compiled version.')



import typing as t
from string import Template
from vernacular.i18nstr import i18nstr
from vernacular.store import Translations


class Translator:

    __slots__ = ('translations', 'default_domain', 'default_language')

    def __init__(self,
                 translations: Translations,
                 default_domain: str = 'messages',
                 default_language: str = 'en_US'):
        self.translations = translations
        self.default_domain = default_domain
        self.default_language = default_language

    def translate(self,
                  phrase: t.Union[i18nstr, str],
                  domain=None,
                  mapping=None,
                  context=None,
                  default=None,
                  target_language=None):

        if not isinstance(phrase, i18nstr):
            phrase = i18nstr(
                phrase,
                domain=domain or self.default_domain,
                mapping=mapping,
                context=context,
                default=default
            )
        else:
            phrase = phrase.replace(
                domain=domain or phrase.domain or self.default_domain,
                mapping=mapping,
                context=context,
                default=default
            )

        target_language = target_language or self.default_language
        translated = None
        if trdomain := self.translations.get(phrase.domain):
            translated = trdomain.lookup(phrase, target_language)
        if translated is None:
            translated = phrase.default
        if translated and '$' in translated and phrase.mapping:
            return Template(translated).safe_substitute(phrase.mapping)

        return translated

    def pluralize(self,
                  singular: str, plural: str, num: int,
                  domain=None,
                  mapping=None,
                  context=None,
                  target_language=None):
        domain = domain or self.default_domain
        target_language = target_language or self.default_language
        translated = None
        if trdomain := self.translations.get(domain):
            translated = trdomain.nlookup(
                singular, plural, num, target_language)
        if translated and '$' in translated and mapping:
            return Template(translated).safe_substitute(mapping)
        return translated



import gettext
import typing as t
from pathlib import Path
import langcodes


class Translation(t.NamedTuple):
    domain: str
    locale: str
    mofile: Path


Language = t.Dict[str, gettext.GNUTranslations]


class Domain(t.Dict[str, Language]):

    domain: str

    def __init__(self, domain: str, *args, **kwargs):
        self.domain = domain
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f'<Translation domain "{self.domain}">'

    def add(self, translation: Translation):
        if translation.domain != self.domain:
            raise ValueError(
                f'{self!r} can only contain translations from '
                f'domain {self.domain}. Got {translation.domain}'
            )
        try:
            tag = langcodes.Language.get(translation.locale)
        except ValueError:
            return None
        if tag.language not in self:
            language = self[tag.language] = {}
        else:
            language = self[tag.language]

        with translation.mofile.open('rb') as fp:
            catalog = gettext.GNUTranslations(fp=fp)

        if tag.territory is not None:
            if None not in language:
                # we are creating a locale in this language but we have
                # no root
                root = language[None] = gettext.GNUTranslations(fp=None)
                # there was no parse. Init the catalog
                root._catalog = {}
                root.plural = catalog.plural
            else:
                root = language[None]
            catalog.add_fallback(root)

        if tag.territory in language:
            language[tag.territory]._catalog.update(catalog._catalog)
        else:
            language[tag.territory] = catalog

    def lookup(self, entry: str, language: str):
        tag = langcodes.Language.get(language)
        if catalogs := self.get(tag.language):
            catalog = catalogs[None]
            if tag.territory:
                catalog = catalogs.get(tag.territory, catalog)
            return catalog.gettext(entry)

    def nlookup(self, singular: str, plural: str, num: int, language: str):
        tag = langcodes.Language.get(language)
        if catalogs := self.get(tag.language):
            catalog = catalogs[None]
            if tag.territory:
                catalog = catalogs.get(tag.territory, catalog)
            return catalog.ngettext(singular, plural, num)


class Translations(t.Dict[str, Domain]):

    def add(self, translation: Translation):
        if translation.domain not in self:
            self[translation.domain] = Domain(translation.domain)
        self[translation.domain].add(translation)



import os
import typing as t
import logging
import contextvars
from pathlib import Path
from importlib.metadata import entry_points
from hamcrest.core.base_matcher import BaseMatcher
from hamcrest import starts_with, not_
from vernacular.store import Translation, Translations
from vernacular.utils import iter_translation_files


Logger = logging.getLogger(__name__)
COMPILE = os.environ.get('COMPILE_MO_FILES', False)
Language = contextvars.ContextVar('language')


def translations(path: Path) -> t.List[Translation]:
    return [
        Translation(domain=mofile.stem, locale=locale, mofile=mofile)
        for locale, mofile in iter_translation_files(
                path, can_compile=COMPILE
        )
    ]


def from_entrypoints(
        restrict: t.Optional[BaseMatcher] = not_(starts_with('test_')),
        key: str = 'vernacular.translation_directory'):
    translations = Translations()
    eps = entry_points()
    if (i18n_defs := eps.select(group=key)) is not None:
        for definition in i18n_defs:
            if restrict and not restrict.matches(definition.name):
                continue

            i18_files = definition.load()
            for translation in i18_files():
                translations.add(translation)
                Logger.debug(
                    f"Loading i18n registration: {translation}."
                )
    return translations


def translations_test_folder():
    return translations(Path(__file__).parent / 'locales')


def clean_translations_test_folder():
    """removes all the .mo - for test purposes
    """
    path = Path(__file__).parent / 'locales'
    for lang in path.iterdir():
        messages = lang / 'LC_MESSAGES'
        for child in messages.iterdir():
            if child.suffix == '.mo':
                child.unlink()



from setuptools import setup


setup(
    name='vernacular',
    install_requires = [
        'frozendict',
        'pyhamcrest',
        'langcodes',
        'polib',
    ],
    extras_require={
        'test': [
            'WebTest',
            'pytest',
        ]
    }
)
